import transaction
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import scoped_session, sessionmaker
from zope.sqlalchemy import ZopeTransactionExtension
from sqlalchemy import Column, String, Integer, UniqueConstraint, and_

DBSession = scoped_session(sessionmaker(extension=ZopeTransactionExtension()))  # initialise a ORM session, ought to be reused across the different modules


class ORMClass(object):
    """
    Declares a parent class for all models which eases querying
    """

    @classmethod
    def query(cls):
        """
        Pass down the class name when using the DBSession.query method and use ModelClass.query() instead of DBSession.query(ModelClass)

        :return: the query result object
        """

        global DBSession

        return DBSession.query(cls)

    @staticmethod
    def add(instance):
        """
        Adds an instance of a model to the database.

        :param instance: the instance to be created in the db.
        """

        global DBSession

        with transaction.manager:
            DBSession.add(instance)

    @staticmethod
    def delete(instance):
        """
        Deletes an instance of a model from the database.

        :param instance: the instance to be deleted from the db.
        """

        global DBSession

        with transaction.manager:
            DBSession.delete(instance)

    @staticmethod
    def replace(old_instance, new_instance):
        """
        Replaces an instance of a model from the database with a new instance.

        :param old_instance: the instance to be replaced from the db.
        :param new_instance: the new instance
        """

        global DBSession

        with transaction.manager:
            DBSession.add(new_instance)
            DBSession.delete(old_instance)

    @classmethod
    def delete_all(cls):
        """
        Deletes all instances of a model from the database.
        """

        global DBSession

        with transaction.manager:
            deleted_rows = DBSession.query(cls).delete()

        return deleted_rows


Base = declarative_base(cls=ORMClass)  # initialise a declarative Base instance to use for the web app models (inherits from the base ORM class defined above)


class ServiceFunctionEndpoint(Base):
    """
    This class defines the main model of the WHOAMI API, declaring the global tags for a specific service function on a specific endpoint.
    """

    __tablename__ = 'sfendpoints'  # table name in the PostgreSQL database

    __table_args__ = (UniqueConstraint('sf_i', 'sf_endpoint', 'sr'),)  # defines a unique constraint across 3 columns - sf_i, sf_endpoint, sr

    uid = Column(Integer, primary_key=True, autoincrement=True, nullable=False)  # a primary key integer field (auto incremented)

    location = Column(String, nullable=False)  # cluster label
    sfc = Column(String, nullable=False)  # service function chain label
    sfc_i = Column(String, nullable=False)  # service function chain instance identifier
    sf = Column(String, nullable=False)  # service function label
    sf_i = Column(String, nullable=False)   # service function identifier (potentially FQDN)
    sf_endpoint = Column(String, nullable=False)  # service function endpoint (potentially IP address)
    sr = Column(String, nullable=False)  # service router ID - service router that connects the VM to FLAME

    @property
    def json(self):
        """
        Converts an instance of a ServiceFunctionEndpoint to JSON format.

        :return: a python dictionary object
        """

        fields = {c.name: getattr(self, c.name) for c in self.__table__.columns}
        fields.pop("uid")

        return fields

    @staticmethod
    def required_columns():
        """
        Returns the required columns for constructing a valid instance.
        :return: a generator object
        """

        return tuple(column.name for column in ServiceFunctionEndpoint.__table__.columns if column.name != "uid")

    @staticmethod
    def constrained_columns():
        """
        :return: the columns that are uniquely identifying an instance of this model.
        """

        return tuple(column.name for column in ServiceFunctionEndpoint.__table_args__[0].columns)

    @staticmethod
    def get(sf_i, sf_endpoint, sr):
        """
        Gets the instance matching the unique constraint or None if not existing.

        :param sf_i: service function instance
        :param sf_endpoint: service function endpoint
        :param sr: service router
        :return: the first object from the result set that matches the unique constraint or None
        """

        return ServiceFunctionEndpoint.query().filter(and_(ServiceFunctionEndpoint.sf_i == sf_i, ServiceFunctionEndpoint.sf_endpoint == sf_endpoint, ServiceFunctionEndpoint.sr == sr)).first()

    @staticmethod
    def exists(sf_i, sf_endpoint, sr):
        """
        Checks if an instance matching the unique constraint exists.

        :param sf_i: service function instance
        :param sf_endpoint: service function endpoint
        :param sr: service router
        :return: True if exists, False otherwise
        """

        return ServiceFunctionEndpoint.get(sf_i, sf_endpoint, sr) is not None
